const gql = require('graphql-tag')

const changeCase = require('change-case')
const inflection = require( 'inflection' )

const yves = require('yves')

const _ = require('lodash')
const globToRegExp = require('glob-to-regexp')

const pkg = require('../package.json')
const fs = require('fs')
const path = require('path')

const { PubSub } = require('graphql-subscriptions')
const { withFilter } = require('apollo-server')
const sift = require('sift')

const adapterSpecialArgs = [
  '_adapter',
  '_collection',
  '_filter',
  '_search',
  '_limit',
  '_skip',
  '_sort',
  '_options',
  '_field',
  '_data',
]

function moduleSetup(moduleOptions,moduleName) {
  const debug = yves.debugger(`${pkg.name.replace(/-/g,':')}:${moduleName.replace(/-/g,':')}`)
  let moduleSchema
  const rootName = moduleOptions.root
  const rootTypes = _.get(moduleOptions,'rootTypes',['query'])

  const pubsub = (rootTypes.indexOf('subscription') >= 0) ? new PubSub() : null

  // if (rootTypes.indexOf('subscription') >= 0) {
  //   setInterval( () => {
  //     pubsub.publish('change_stream', { juke: { id: "abc",date:new Date() }});        
  //     debug ('published') 
  //   },3000)
  // }

  if (moduleOptions.enabled && moduleOptions.schema) {
    debug('input schema %y', moduleOptions.schema)
    moduleSchema = require(moduleOptions.schema)
  }
  const INTROSPECT = 'introspect'
  const TYPE_INTROSPECT = `${rootName}__${changeCase.pascalCase(INTROSPECT)}`

  const resolvers = {}

  if (rootTypes.indexOf('query') >= 0) {
    resolvers['Query'] = {
      [rootName.toLowerCase()]: (_source, args, context ) => {
        if (context.dataSources[moduleName]) {
          context.dataSources[moduleName].schema = moduleSchema
          return context.dataSources[moduleName].access(args.token, 'query', context)
        }
      },
    }
    resolvers[`${rootName}__Query`] = {}
  }

  if (rootTypes.indexOf('mutation') >= 0) {
    resolvers['Mutation'] = {
      [rootName.toLowerCase()]: (_source, args, context ) => {
        if (context.dataSources[moduleName]) {
          context.dataSources[moduleName].schema = moduleSchema
          return context.dataSources[moduleName].access(args.token, 'mutation', context)
        }
      },
    }
    resolvers[`${rootName}__Mutation`] = { update: () => { return {} }, insert: () => { return {} }, delete: () => { return {} } }
    resolvers[`${rootName}__Mutation_Update`] =  {}
    resolvers[`${rootName}__Mutation_Insert`] = {}
    resolvers[`${rootName}__Mutation_Delete`] = {}
  }

  if (rootTypes.indexOf('subscription') >= 0) {
    resolvers['Subscription'] = {
      [rootName.toLowerCase()]: {
        subscribe: withFilter(
          () => pubsub.asyncIterator(`${rootName.toLowerCase()}__change`), (change, args, context) => {
            let filter = (args.filter && typeof args.filter == 'object' && Object.keys(args.filter).length) ? args.filter : null
            // debug('markers root %s %y',rootName.toLowerCase(),args.markers[rootName.toLowerCase()])
            if (args.markers && args.markers[rootName.toLowerCase()]) {
              // debug('markers %y',args.markers)
              const and = [{collection:{$in: Object.keys(_.get(moduleSchema,'collections',{}))}}]
              const or = []
              if (filter) or.push(filter)
              for (let collection in args.markers[rootName.toLowerCase()]) {
                const ids = args.markers[rootName.toLowerCase()][collection].map( id => ((typeof id == 'string' && (id.indexOf('*')>=0 || id.indexOf('?')>=0 || id.indexOf('[')>=0 || id.indexOf(']')>=0)) ? globToRegExp(id) : id) )

                or.push({collection,'document._id':{$in: ids}})
              }
              if (or.length) and.push({$or: or})//.$or = or // filter = {$or: or}
              filter = {$and: and}
              // if (or.length) {
              //   if (filter) {
              //     const and = []
              //     and.push(filter)
              //     and.push({$or: or})
              //     filter = { $and: and }
              //   } else {
              //     filter = {$or: or}
              //   }
              // }
            }
            //debug('sift change %y filter %y pass %y',change[rootName.toLowerCase()],filter,(!filter || sift(filter)(change[rootName.toLowerCase()])))
            return !!change[rootName.toLowerCase()] && (!filter || sift(filter)(change[rootName.toLowerCase()]) )
          }
        )
      }
    }
  }


  function prefix(name) 
  {
    if (!name) return ''
    const l=name.toLowerCase().trim()
    if (l==='int' || l==='float' || l === 'string' || l==='boolean' || l === 'id' || l==='object' || l==='date') return name
    return `${rootName}_${name}`
  }

  let gqlSchema = ''
  if (rootTypes.indexOf('query') >= 0) {
    gqlSchema += `
extend type Query {
  "${rootName} accessor"
  ${rootName.toLowerCase()}${moduleOptions.authentication && moduleOptions.authentication.enabled && moduleOptions.authentication.type === 'token' && _.get(moduleOptions,'authentication.realms',['query','mutation','subscription']).indexOf('query') >= 0 ? '(token: String!)':''}: ${rootName}__Query
}

type ${rootName}__Query {
${adaptorQueryMembers()}}

type ${TYPE_INTROSPECT} {

  "MongoDB collection"
  collection: String

  "MongoDB filter"
  filter: Object

  "MongoDB skip option"
  skip: Int

  "MongoDB limit option"
  limit: Int

  "MongoDB sort option"
  sort: Object

  "MongoDB options"
  options: Object

  "Result item count"
  count: Int

  "Total item count, not skipped nor limited"
  total: Int

  "MongoDB result"
  data: [ Object ]

  "MongoDB explanation (show vital index information)"
  explain: Object
}
`
  }
  if (rootTypes.indexOf('mutation') >= 0) {
    gqlSchema += `
extend type Mutation {
  "${rootName} accessor"
  ${rootName.toLowerCase()}${moduleOptions.authentication && moduleOptions.authentication.enabled && moduleOptions.authentication.type === 'token'  && _.get(moduleOptions,'authentication.realms',['query','mutation','subscription']).indexOf('mutation') >= 0 ? '(token: String!)':''}: ${rootName}__Mutation
}

type ${rootName}__Mutation {
  update: ${rootName}__Mutation_Update
  insert: ${rootName}__Mutation_Insert
  delete: ${rootName}__Mutation_Delete
}

type ${rootName}__Mutation_Update {
${adaptorMutationUpdateMembers()}}

type ${rootName}__Mutation_Insert {
${adaptorMutationInsertMembers()}}
  
type ${rootName}__Mutation_Delete {
${adaptorMutationDeleteMembers()}}

${adaptorInputTypes()}
`
  }

  if (rootTypes.indexOf('subscription') >= 0) {
    gqlSchema += `
extend type Subscription {
  ${rootName.toLowerCase()}(filter: Object, markers: Object): ${rootName}__Change
}

type ${rootName}__Change {
  operation: String
  collection: String
  date: Date
  document: Object
  fields: Object
}

`
  }

  
  // debug('graphql %y',gqlSchema)

  
  function abstractSchemaType2graphQLtype(field,name) {
    switch (field.type) {
    case 'string':
      if (name == '_id' || _.get(field,'reference') || _.get(field,'references') || _.get(field,'member.reference') || _.get(field,'member.references') ) {
        return 'ID'
      } else {
        return 'String'
      }
    case 'integer':
      return 'Int'
    case 'float':
      return 'Float'
    case 'boolean':
      return 'Boolean'
    case 'date':
      return 'Date'
    case 'object':
      return 'Object'
    case 'array':
      return field.member ? `[ ${abstractSchemaType2graphQLtype(field.member,name)} ]` : '[ String ]'            
    }
  }

  function makeParameterSpec(purpose, collection, requiredFields, plural) {
    let result=''
    if ( !plural ) {
      const schema = _.get(moduleSchema,`collections.${collection}.schema`)
      if (schema) {
        for (let field in schema) {
          const gType = abstractSchemaType2graphQLtype(schema[field],field)
          result += ` "Filter field '${field}'" ${field}:${gType}`
          // if (schema[field].type == 'string') {
          //   result += ` "Filter field '${field}'" ${field}:String`
          // } else if (schema[field].type == 'integer') {
          //   result += ` "Filter field '${field}'" ${field}:Int`
          // } else if (schema[field].type == 'float') {
          //   result += ` "Filter field '${field}'" ${field}:Float`
          // } else if (schema[field].type == 'boolean') {
          //   result += ` "Filter field '${field}'" ${field}:Boolean`
          // }
          if (requiredFields && requiredFields.indexOf(field)>=0) {
            result += '!'
          }
        }
      }
      const textIndex = _.get(moduleSchema,`indexes.${collection}`)
      result += ' "Adds an extra MongoDB style filter" _filter: Object'
      if (textIndex) result += ' "Searches the text fields for matches" _search:String'
      result += ' "Limits the number of results, defaults to ${moduleOptions.limit}" _limit: Int'
      result += ' "Skips number of results, defaults to 0" _skip: Int, "Sort instruction MongoDB style" _sort: Object'
      result += ' "Extra MongoDB options" _options: Object'
    }
    switch (purpose) {
    case 'query':
      break
    case 'update':
      result += ' "Data of changes" _data: Object!'
      break  
    case 'insert':
      if (plural) {
        result += ` "Data to be inserted" _data: [ ${rootName}__Input__${inflection.singularize(collection).replace(/\W/g,'')} ]!`
      }
      break  
    case 'delete':
      if (plural) {
        result += ` "Data to be deleted" _data: [ ${rootName}__Input__${inflection.singularize(collection).replace(/\W/g,'')} ]!`
      }
      break  
    }
    return result
  }


  function graphQLmap(schema, container, description, path, title, level = 0) {
    const typeEntries = []
    const unionEntries = []

    for (let entry in schema[container]) {
      const type = schema[container][entry].type
      const oSchema = schema[container][entry].schema

      if ((type == 'collection' || type == 'object') && oSchema && Object.keys(oSchema).length) {
        const mTitle = changeCase.pascalCase( (title ? `${title} ` : '') + inflection.singularize(entry) )
        graphQLmap(schema[container][entry],'schema', schema[container][entry].description, path ? `${path}.${entry}` : entry, mTitle, level + 1)
        if (type == 'object') typeEntries.push(`  ${entry.replace(/\W/g,'')}: ${prefix(mTitle)}`)
      } else if (type == 'array') {
        const memberType = schema[container][entry].member.type
        const memberSchema = schema[container][entry].member.schema
        let mTitle = changeCase.pascalCase(memberType == 'object' && (!memberSchema || Object.keys(memberSchema).length == 0) ? 'object' : memberType.replace('integer','int'))
        if (memberType == 'object' && memberSchema && Object.keys(memberSchema).length) {
          mTitle = changeCase.pascalCase( (title ? `${title} ` : '') + inflection.singularize(entry) )
          graphQLmap(schema[container][entry].member,'schema', null,path ? `${path}.${entry}` : entry, mTitle, level + 1)
        }
        const reference = schema[container][entry].member.reference
        if (reference) {
          if (!resolvers[prefix(title)]) resolvers[prefix(title)] = {}
          resolvers[prefix(title)][entry] = (source, args, context) => {
            if (context.dataSources[moduleName]) {
              if (context.dataloader && _.get(moduleOptions,'dataloader.enabled')) {
                return context.dataSources[moduleName].getMany(source[entry],reference,makeFilterWithArgs(args,reference,moduleSchema),makeOptionsWithArgs(args),context)
              } else {
                const filter = context.dataSources[moduleName].mergeFilters({_id:{$in:source[entry]}},makeFilterWithArgs(args,reference,moduleSchema))
                return context.dataSources[moduleName].query(reference,filter,makeOptionsWithArgs(args),context)
              }
            }
          }
          typeEntries.push(`  ${entry.replace(/\W/g,'')}(${makeParameterSpec('query',reference)}): [ ${prefix(changeCase.pascalCase(inflection.singularize(reference)))} ]`)

          if (moduleOptions.introspect.enabled) {
            resolvers[prefix(title)][`${INTROSPECT}_${entry}`] = (source, args, context, info) => {
              const filter = context.dataSources[moduleName].mergeFilters({_id:{$in:source[entry]}},makeFilterWithArgs(args,reference,moduleSchema))
              if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(reference,filter, makeOptionsWithArgs(args), context, info)
            }
            const IAentry = `  ${INTROSPECT}_${entry}(${makeParameterSpec('query',reference)}): ${TYPE_INTROSPECT}`
            if (typeEntries.indexOf(IAentry)<0) typeEntries.push(IAentry)
          }
        } else {
          const references = schema[container][entry].member.references
          if (references) {

            const unionName = `${prefix(title)}_${entry.replace(/\W/g,'')}`
            unionEntries.push(`union ${unionName} = ${references.map( reference => prefix(changeCase.pascalCase(inflection.singularize(reference)))).join(' | ') }`)

            if (!resolvers[prefix(title)]) resolvers[prefix(title)] = {}

            resolvers[prefix(title)][entry] = (source, args, context, info) => {
              // See https://stackoverflow.com/a/46313828
              // debug('fieldNodes %y',_.get(info,'fieldNodes[0].selectionSet.selections',[]))
              const requestedFields = _.get(info,'fieldNodes[0].selectionSet.selections',[])
                .map(s => _.get(s,'name.value'))
                .filter( e => !!e )

              const requestedCollections = _.get(info,'fieldNodes[0].selectionSet.selections',[])
                .map(s => _.get(s,'typeCondition.name.value'))
                .filter( e => !!e )
                .map( type => inflection.pluralize(type.substr(moduleName.length+1).toLowerCase()) )  // reverse map GraphQL Type to mongo collection name

              // debug('references: references %y requestedCollections %y requestedFields %y',references,requestedCollections,requestedFields)
              if (context.dataSources[moduleName]) {
                if (Array.isArray(source[entry]) && source[entry].length) {
                  const subQueries = references
                    .filter( reference => (requestedFields && requestedFields.length) || !requestedCollections || !requestedCollections.length || requestedCollections.indexOf(reference)>=0)
                    .map( reference => {
                      if (context.dataloader && _.get(moduleOptions,'dataloader.enabled')) {
                        return context.dataSources[moduleName].getMany(source[entry],reference,makeFilterWithArgs(args,reference,moduleSchema),makeOptionsWithArgs(args),context)
                      } else {
                        const filter = context.dataSources[moduleName].mergeFilters({_id:{$in:source[entry]}},makeFilterWithArgs(args,reference,moduleSchema))
                        return context.dataSources[moduleName].query(reference,filter,makeOptionsWithArgs(args),context)
                      }
                    })

                  return Promise.all(subQueries).then( resultArray => {
                    const result = [].concat.apply([], resultArray).filter( r => !!r )

                    const itemsByIds = _.keyBy( result, '_id')
                    return source[entry].map(id => (itemsByIds && itemsByIds[id]) || null).filter( r => !!r )
                  })
                } else {
                  return []
                }
              }
            }

            typeEntries.push(`  ${entry.replace(/\W/g,'')}(${makeParameterSpec('query',references)}): [ ${ unionName } ]`)

            if (moduleOptions.introspect.enabled) {
              resolvers[prefix(title)][`${INTROSPECT}_${entry.replace(/\W/g,'')}`] = (source, args, context, info) => {
                const filter = context.dataSources[moduleName].mergeFilters({_id:{$in:source[entry]}},makeFilterWithArgs(args,null,moduleSchema))
                if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspects(references, filter, makeOptionsWithArgs(args), context, info)
              }
              const IAentry = `  ${INTROSPECT}_${entry.replace(/\W/g,'')}(${makeParameterSpec('query',references)}): [ ${TYPE_INTROSPECT} ]`
              if (typeEntries.indexOf(IAentry)<0) typeEntries.push(IAentry)
            }
  
            resolvers[unionName] = {
              __resolveType(entry) {
                return `${changeCase.pascalCase(moduleName)}_${changeCase.pascalCase(inflection.singularize(entry.__collection))}`
              }
            }

          } else {
            typeEntries.push(`  ${entry.replace(/\W/g,'')}: [ ${prefix(mTitle)} ]`)
          }
        }
      } else {
        const reference = (schema[container][entry].reference && moduleSchema.collections[schema[container][entry].reference]) ? schema[container][entry].reference : null
        if (reference) {
          const refType = changeCase.pascalCase(inflection.singularize(reference))

          if (!resolvers[prefix(title)]) resolvers[prefix(title)]={}

          resolvers[prefix(title)][entry.replace(/\W/g,'')] = (source, args, context) => {
            if (context.dataSources[moduleName]) {
              if (context.dataloader && _.get(moduleOptions,'dataloader.enabled')) {
                return context.dataSources[moduleName].getSingle(source[entry], reference, makeFilterWithArgs(args,reference,moduleSchema), makeOptionsWithArgs(args),context)
              } else {
                const filter = context.dataSources[moduleName].mergeFilters({_id:source[entry]},makeFilterWithArgs(args,reference,moduleSchema))
                return context.dataSources[moduleName].query(reference, filter, makeOptionsWithArgs(args),context)
                  .then( result => result && Array.isArray(result) && result.length ? result[0] : result)
              }
            }
          }
          typeEntries.push(`  ${entry.replace(/\W/g,'')}: ${prefix(refType)}`)

          if (moduleOptions.introspect.enabled) {
            resolvers[prefix(title)][`${INTROSPECT}_${entry.replace(/\W/g,'')}`] = (source, args, context, info) => {
              const filter = context.dataSources[moduleName].mergeFilters({_id:source[entry]},makeFilterWithArgs(args,reference,moduleSchema))
              if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(reference, filter, makeOptionsWithArgs(args), context, info)
            }
            const IAentry = `  ${INTROSPECT}_${entry.replace(/\W/g,'')}(${makeParameterSpec('query',reference)}): ${TYPE_INTROSPECT}`
            if (typeEntries.indexOf(IAentry)<0) typeEntries.push(IAentry)
          }
        } else {
          const oSchema = schema[container][entry].schema
          typeEntries.push(`  ${entry.replace(/\W/g,'')}: ${prefix(changeCase.pascalCase(type == 'object' && (!oSchema || Object.keys(oSchema).length == 0) ? 'object' : type.replace('integer','int')))}`)
        }
      }
    }




    if (level == 1) {
      for (let refEntry in moduleSchema.collections) {
        for (let field in moduleSchema.collections[refEntry].schema) {
          const type = moduleSchema.collections[refEntry].schema[field].type
          const reference = moduleSchema.collections[refEntry].schema[field].reference || (type == 'array' && moduleSchema.collections[refEntry].schema[field].member && moduleSchema.collections[refEntry].schema[field].member.reference)
          if (reference == path && moduleSchema.collections[reference]) {
            const reverse = moduleSchema.collections[refEntry].schema[field].reverse || (type == 'array' && moduleSchema.collections[refEntry].schema[field].member && moduleSchema.collections[refEntry].schema[field].member.reverse)
            const mainType = changeCase.pascalCase(inflection.singularize(path))
            if (!resolvers[prefix(mainType)]) resolvers[prefix(mainType)]={}
            let virtField = reverse || inflection.pluralize(refEntry).replace(/\W/g,'')
            if (resolvers[prefix(mainType)][virtField]) virtField=field.toLowerCase()+'_'+virtField

            resolvers[prefix(mainType)][virtField] = (source, args, context) => {
              const filter = context.dataSources[moduleName].mergeFilters({[field]:source._id},makeFilterWithArgs(args,refEntry,moduleSchema))
              return context.dataSources[moduleName].query(refEntry,filter,makeOptionsWithArgs(args),context)
            }
            const gentry = `  ${virtField}(${makeParameterSpec('query',refEntry)}): [ ${prefix(changeCase.pascalCase(inflection.singularize(refEntry)))} ]`
            if (typeEntries.indexOf(gentry)<0) typeEntries.push(gentry)
            if (moduleOptions.introspect.enabled) {
              resolvers[prefix(mainType)][`${INTROSPECT}_${virtField}`] = (source, args, context, info) => {
                const filter = context.dataSources[moduleName].mergeFilters({[field]:source._id},makeFilterWithArgs(args,refEntry,moduleSchema))
                if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(refEntry,filter, makeOptionsWithArgs(args), context, info)
              }
              const IAentry = `  ${INTROSPECT}_${virtField}(${makeParameterSpec('query',refEntry)}): ${TYPE_INTROSPECT}`
              if (typeEntries.indexOf(IAentry)<0) typeEntries.push(IAentry)
            }
          }
        }
      }
    }


    if (typeEntries.length) {
      const pat = new RegExp(`^\\s*${INTROSPECT}_`, '')

      const unions = unionEntries.join('\n')
      if (unions) gqlSchema += unions+'\n\n'

      if (description) gqlSchema += `"${description}"\n`

      gqlSchema += `type ${prefix(title)} {\n${typeEntries.sort((a, b) => {
        const m_a = a.match(pat)
        const m_b = b.match(pat)
        if (m_a != m_b) {
          if (m_a) return 1
          if (m_b) return -1
        }
        return 0
      }).join('\n')}\n}\n\n`
    }
  }


  if (rootTypes.indexOf('query') >= 0) {
    if (moduleSchema) graphQLmap(moduleSchema,'collections')
  }

  function adaptorQueryMembers() {
    let result=''
    let result2=''
    if (moduleSchema) {
      const collections = Object.keys(moduleSchema.collections)

      collections.forEach( collection => {
        result += `  ${inflection.singularize(collection).replace(/\W/g,'')}(${makeParameterSpec('query',collection)}): ${prefix(changeCase.pascalCase( inflection.singularize(collection) ).replace(/\W/g,''))}\n`
        result += `  ${inflection.pluralize(collection).replace(/\W/g,'')}(${makeParameterSpec('query',collection)}): [ ${prefix(changeCase.pascalCase( inflection.singularize(collection) ).replace(/\W/g,''))} ]\n`

        if (moduleOptions.introspect.enabled) {
          result2 += `  ${INTROSPECT}_${inflection.singularize(collection).replace(/\W/g,'')}(${makeParameterSpec('query',collection)}): ${TYPE_INTROSPECT}\n`
          result2 += `  ${INTROSPECT}_${inflection.pluralize(collection).replace(/\W/g,'')}(${makeParameterSpec('query',collection)}): ${TYPE_INTROSPECT}\n`
        }
      })
    }
    return result+result2
  }


  function adaptorInputTypesSchema(schema)
  {
    let result=''
    if (schema) {
      for (let field in schema) {
        const gType = abstractSchemaType2graphQLtype(schema[field],field) // schema[field].schema ? `{\n${adaptorInputTypesSchema(schema[field].schema)}}\n` : abstractSchemaType2graphQLtype(schema[field],field)
        result += `  ${field}: ${gType}\n`
      }
    }    
    return result
  }

  function adaptorInputTypes() {
    let result=''
    if (moduleSchema) {
      const collections = Object.keys(moduleSchema.collections)

      collections.forEach( collection => {
        result += `input ${rootName}__Input__${inflection.singularize(collection).replace(/\W/g,'')} {\n`

        const schema = _.get(moduleSchema,`collections.${collection}.schema`)
        result += adaptorInputTypesSchema(schema)
        // if (0 && schema) {
        //   for (let field in schema) {
        //     const gType = abstractSchemaType2graphQLtype(schema[field],field)
        //     result += `  ${field}: ${gType}\n`
        //   }
        // }    
        result += '}\n\n'
//        debug('gql %y',result)
      })
    }
    return result
  }


  function adaptorMutationUpdateMembers() {
    let result=''
    if (moduleSchema) {
      const collections = Object.keys(moduleSchema.collections)

      collections.forEach( collection => {
        result += `  ${inflection.singularize(collection).replace(/\W/g,'')}(${makeParameterSpec('update',collection)}): ${prefix(changeCase.pascalCase( inflection.singularize(collection) ).replace(/\W/g,''))}\n`
      })
    }
    return result
  }

  function adaptorMutationInsertMembers() {
    let result=''
    if (moduleSchema) {
      const collections = Object.keys(moduleSchema.collections)

      collections.forEach( collection => {
        result += `  ${inflection.singularize(collection).replace(/\W/g,'')}(${makeParameterSpec('insert',collection) }): ${prefix(changeCase.pascalCase( inflection.singularize(collection) ).replace(/\W/g,''))}\n`
        result += `  ${inflection.pluralize(collection).replace(/\W/g,'')}(${makeParameterSpec('insert',collection,[],true) }): [ ${prefix(changeCase.pascalCase( inflection.singularize(collection) ).replace(/\W/g,''))} ]\n`
      })
    }
    return result
  }

  function adaptorMutationDeleteMembers() {
    let result=''
    if (moduleSchema) {
      const collections = Object.keys(moduleSchema.collections)

      collections.forEach( collection => {
        result += `  ${inflection.singularize(collection).replace(/\W/g,'')}(${makeParameterSpec('delete',collection)}): ${prefix(changeCase.pascalCase( inflection.singularize(collection) ).replace(/\W/g,''))}\n`
        result += `  ${inflection.pluralize(collection).replace(/\W/g,'')}(${makeParameterSpec('delete',collection,[],true)}): [ ${prefix(changeCase.pascalCase( inflection.singularize(collection) ).replace(/\W/g,''))} ]\n`
      })
    }
    return result
  }

  if (moduleOptions.writeGraphQLfile) {
    const filename=path.join(path.dirname(moduleOptions.schema),`${rootName.toLowerCase()}.graphql`)
    fs.writeFileSync(filename, gqlSchema)
    debug('output schema %y',filename)
  }

  const schema = moduleSchema ? gql([gqlSchema]) : null


  if (moduleSchema) {
    const collections = Object.keys(moduleSchema.collections)
    collections.forEach( collection => {

      // Subscription: {
      //   somethingChanged: {
      //     subscribe: () => pubsub.asyncIterator(SOMETHING_CHANGED_TOPIC),
      //   },
      // },
      // resolvers[`${rootName}__Subscription`]['changed'] = {
      //   subscribe: () => {
      //     debug('subscribe')
      //     return pubsub.asyncIterator('something_changed')
      //   },
      // }

      // resolvers[`${rootName}__Subscription`][inflection.singularize(collection).replace(/\W/g,'')] = (_source, args, context) => {
      //   if (context.dataSources[moduleName]) {
      //     if (context.dataloader && _.get(moduleOptions,'dataloader.enabled') && args._id) {
      //       const filter=makeFilterWithArgs(args,collection,moduleSchema)
      //       delete(filter._id)
      //       return context.dataSources[moduleName].getSingle(args._id, collection, filter, makeOptionsWithArgs(args), context)
      //     } else {
      //       return context.dataSources[moduleName].query(collection, makeFilterWithArgs(args,collection,moduleSchema), makeOptionsWithArgs(args,{limit:1}), context)
      //         .then( result => {
      //           return result && Array.isArray(result) && result.length ? result[0] : result
      //         })
      //     }
      //   }
      // }


      if (rootTypes.indexOf('mutation') >= 0) {
        resolvers[`${rootName}__Mutation_Update`][inflection.singularize(collection).replace(/\W/g,'')] = (_source, args, context) => {
          if (context.dataSources[moduleName]) {
            return context.dataSources[moduleName].update(collection, makeFilterWithArgs(args,collection,moduleSchema), {$set: args._data}, makeOptionsWithArgs(args,{upsert:true}), context).then( r => {
              if (context.dataSources[moduleName]) {
                // debug('Update: %y', r)
                if (_.get(r,'lastErrorObject.n') != 1 || !_.get(r,'lastErrorObject.updatedExisting')) {
                  throw new Error('Not updated a single existing item.')
                }
                return context.dataSources[moduleName].query(collection,makeFilterWithArgs(args,collection,moduleSchema), makeOptionsWithArgs(args), context)
                  .then( result => result && Array.isArray(result) && result.length ? result[0] : result)
              }
            })
          }
        }
        
        resolvers[`${rootName}__Mutation_Insert`][inflection.singularize(collection).replace(/\W/g,'')] = (_source, args, context) => {
          if (context.dataSources[moduleName]) {
            return context.dataSources[moduleName].insert(collection, args._data, args, context).then( r => {
              if (context.dataSources[moduleName]) {
                // debug('Insert: %y', r)
                if (context.dataloader && _.get(moduleOptions,'dataloader.enabled') && args._id) {
                  const filter=makeFilterWithArgs(args,collection,moduleSchema)
                  delete(filter._id)
                  return context.dataSources[moduleName].getSingle(r.insertedId, collection, null, makeOptionsWithArgs(args), context)
                } else {
                  return context.dataSources[moduleName].query(collection, {_id: r.insertedId}, makeOptionsWithArgs(args,{limit:1}), context)
                    .then( result => {
                      return result && Array.isArray(result) && result.length ? result[0] : result
                    })
                }
              }
            })
          }
        }
        
        resolvers[`${rootName}__Mutation_Insert`][inflection.pluralize(collection).replace(/\W/g,'')] = (_source, args, context) => {
          if (context.dataSources[moduleName]) {
            const promises = args._data.map( data => context.dataSources[moduleName].insert(collection, data, args, context).then( r => {
              // debug('Insert: %y', r)
              if (context.dataloader && _.get(moduleOptions,'dataloader.enabled') && args._id) {
                const filter=makeFilterWithArgs(args,collection,moduleSchema)
                delete(filter._id)
                return context.dataSources[moduleName].getSingle(r.insertedId, collection, null, makeOptionsWithArgs(args), context)
              } else {
                return context.dataSources[moduleName].query(collection,{_id: r.insertedId}, makeOptionsWithArgs(args,{limit:1}), context)
                  .then( result => result && Array.isArray(result) && result.length ? result[0] : result)
              }
            }))
            return Promise.all(promises)
          }
        }
        
        resolvers[`${rootName}__Mutation_Delete`][inflection.singularize(collection).replace(/\W/g,'')] = (_source, args, context) => {
          if (context.dataSources[moduleName]) {
            const filter = makeFilterWithArgs(args,collection,moduleSchema)
            return context.dataSources[moduleName].delete(collection, filter, args, context).then( (r) => {
              //debug('Delete: %y', r)
              if (_.get(r,'result.n')) {
                return filter
              } else {
                return null
              }
            })
          }
        }
        
        resolvers[`${rootName}__Mutation_Delete`][inflection.pluralize(collection).replace(/\W/g,'')] = (_source, args, context) => {
          if (context.dataSources[moduleName]) {
            const promises = args._data.map( data => context.dataSources[moduleName].delete(collection, data, args, context).then( r => {
              //debug('Delete: %y', r)
              if (_.get(r,'deletedCount')) {
                return data
              } else {
                return null
              }
            }))
            return Promise.all(promises)
          }
        }
      }
      
      if (rootTypes.indexOf('query') >= 0) {
        resolvers[`${rootName}__Query`][inflection.singularize(collection).replace(/\W/g,'')] = (_source, args, context) => {
          if (context.dataSources[moduleName]) {
            if (context.dataloader && _.get(moduleOptions,'dataloader.enabled') && args._id) {
              const filter=makeFilterWithArgs(args,collection,moduleSchema)
              delete(filter._id)
              return context.dataSources[moduleName].getSingle(args._id, collection, filter, makeOptionsWithArgs(args), context)
            } else {
              return context.dataSources[moduleName].query(collection, makeFilterWithArgs(args,collection,moduleSchema), makeOptionsWithArgs(args,{limit:1}), context)
                .then( result => {
                  return result && Array.isArray(result) && result.length ? result[0] : result
                })
            }
          }
        }
        resolvers[`${rootName}__Query`][inflection.pluralize(collection).replace(/\W/g,'')] = (_source, args, context) => {
          if (context.dataSources[moduleName]) return context.dataSources[moduleName].query(collection,makeFilterWithArgs(args,collection,moduleSchema), makeOptionsWithArgs(args), context)
        }

        if (moduleOptions.introspect.enabled) {
          resolvers[`${rootName}__Query`][`${INTROSPECT}_${inflection.singularize(collection).replace(/\W/g,'')}`] = (_source, args, context, info) => {
            if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(collection, makeFilterWithArgs(args,collection,moduleSchema), makeOptionsWithArgs(args,{limit:1}), context, info)
          }
          resolvers[`${rootName}__Query`][`${INTROSPECT}_${inflection.pluralize(collection).replace(/\W/g,'')}`] = (_source, args, context, info) => {
            if (context.dataSources[moduleName]) return context.dataSources[moduleName].introspect(collection, makeFilterWithArgs(args,collection,moduleSchema), makeOptionsWithArgs(args), context, info)
          }
        }
      }
    })
  }

  return moduleSchema ? {
    schema,
    resolvers,
    rootTypes: {
      query: rootTypes.indexOf('query') >= 0,
      mutation: rootTypes.indexOf('mutation') >= 0,
      subscription: rootTypes.indexOf('subscription') >= 0,
    },
    pubsub: (rootTypes.indexOf('subscription') >= 0) ? pubsub : null,
  } : null

}





function makeFilterWithArgs(args,collection,schema) {
  if (args && !collection && args._collection) collection = args._collection

  function sensibleObject(filter) {
    return filter && (typeof filter == 'object') && (Object.keys(filter).length > 0)
  }

  function mergeFilters(...queries) {
    let result
    const trueQueries = queries.filter( query => sensibleObject(query) )
    if (trueQueries.length == 1) {
      result = trueQueries[0]
    } else if (trueQueries.length > 1) {
      result = { $and: trueQueries }
    } else {
      result = {}
    }
    return result
  }

  let $text
  if (args._search) {
    $text={$text: { $search: args._search, $language:'none' } }
  }
  const fieldsFilter={}
  const fieldArgs = Object.keys(args).filter( arg => adapterSpecialArgs.indexOf(arg)<0)
  fieldArgs.forEach(arg => {
    
    //JJR * ? [ ]
    if (_.get(schema,`collections.${collection}.schema.${arg}.type`)=='string' && typeof args[arg] == 'string' && (args[arg].indexOf('*')>=0 || args[arg].indexOf('?')>=0 || args[arg].indexOf('[')>=0 || args[arg].indexOf(']')>=0)) {
      fieldsFilter[arg]={ $regex: globToRegExp(args[arg]), $options: 'i' }
    } else {
      fieldsFilter[arg]=args[arg]
    }
  })
  return mergeFilters(args._filter,fieldsFilter,$text)
}


function makeOptionsWithArgs(args,extra) {
  function cleanObject(obj) {
    if (typeof(obj) == 'object') Object.keys(obj).forEach(function(key) {if (obj[key] === undefined) delete obj[key]} )
    return obj || {}
  }

  function makeOptions(options,obj,extra) {
    return Object.assign( {}, cleanObject(options), cleanObject(obj), cleanObject(extra))
  }

  return makeOptions(args._options,{limit:args._limit,skip:args._skip,sort:args._sort},extra)
}

module.exports = {
  moduleSetup,
  adapterSpecialArgs,
  makeFilterWithArgs,
  makeOptionsWithArgs,
}
